# https://github.com/facebookresearch/esm/blob/main/esm/inverse_folding/multichain_util.py
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import List

import biotite.structure
import numpy as np
# import torch
from biotite.sequence import ProteinSequence
# from biotite.structure.residues import get_residues
from biotite.structure.residues import get_residue_starts

from antifold.esm.inverse_folding.util import (get_encoder_output,
                                               get_sequence_loss,
                                               load_structure)


def get_atom_coords_residuewise(atoms: List[str], struct: biotite.structure.AtomArray):
    """
    Example for atoms argument: ["N", "CA", "C"]
    """

    def filterfn(s, axis=None):
        filters = np.stack([s.atom_name == name for name in atoms], axis=1)
        sum = filters.sum(0)
        if not np.all(sum <= np.ones(filters.shape[1])):
            raise RuntimeError("structure has multiple atoms with same name")
        index = filters.argmax(0)
        coords = s[index].coord
        coords[sum == 0] = float("nan")
        return coords

    return biotite.structure.apply_residue_wise(struct, struct, filterfn)


def get_residues_imgt(array):
    """Get the residue IDs+insertion code (IMGT) and names of an atom array (stack)"""
    starts = get_residue_starts(array)

    # Position + insertion code
    res_pos = array.res_id[starts]
    res_ins = array.ins_code[starts]
    res_chains = array.chain_id[starts]
    res_posinschain = np.array(
        list(
            [
                f"{pos}{ins}{chain}"
                for pos, ins, chain in zip(res_pos, res_ins, res_chains)
            ]
        )
    )

    # Residue name, 3-letter code
    res_name = array.res_name[starts]

    return res_pos, res_posinschain, res_name


def extract_coords_from_structure(structure: biotite.structure.AtomArray):
    """
    Args:
        structure: An instance of biotite AtomArray
    Returns:
        Tuple (coords, seq, res_posinschain)
            - coords (array) is L x 3 x 3 for N, CA, C coordinates
            - seq (str) is the extracted sequence
            - res_posinschain (str) is the residue position IDs with insertion code (e.g. IMGT numbering)
    """
    coords = get_atom_coords_residuewise(["N", "CA", "C"], structure)
    res_pos, res_posinschain, res_3letter = get_residues_imgt(structure)
    seq = "".join([ProteinSequence.convert_letter_3to1(r) for r in res_3letter])
    return coords, seq, res_pos, res_posinschain


def extract_coords_from_complex(structure: biotite.structure.AtomArray):
    """
    Args:
        structure: biotite AtomArray
    Returns:
        Tuple (coords_list, seq_list)
        - coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
          coordinates representing the backbone of each chain
        - seqs: Dictionary mapping chain ids to native sequences of each chain
    """
    coords = {}
    seqs = {}
    positions = {}
    positions_ins = {}
    all_chains = biotite.structure.get_chains(structure)
    for chain_id in all_chains:
        chain = structure[structure.chain_id == chain_id]
        (
            coords[chain_id],
            seqs[chain_id],
            positions[chain_id],
            positions_ins[chain_id],
        ) = extract_coords_from_structure(chain)
    return coords, seqs, positions, positions_ins


def load_complex_coords(fpath, chains):
    """
    Args:
        fpath: filepath to either pdb or cif file
        chains: the chain ids (the order matters for autoregressive model)
    Returns:
        Tuple (coords_list, seq_list)
        - coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
          coordinates representing the backbone of each chain
        - seqs: Dictionary mapping chain ids to native sequences of each chain
    """
    assert os.path.exists(fpath)

    structure = load_structure(fpath, chains)
    return extract_coords_from_complex(structure)


def _concatenate_coords(coords, target_chain_id, padding_length=10):
    """
    Args:
        coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
            coordinates representing the backbone of each chain
        target_chain_id: The chain id to sample sequences for
        padding_length: Length of padding between concatenated chains
    Returns:
        Tuple (coords, seq)
            - coords is an L x 3 x 3 array for N, CA, C coordinates, a
              concatenation of the chains with padding in between
            - seq is the extracted sequence, with padding tokens inserted
              between the concatenated chains
    """
    pad_coords = np.full((padding_length, 3, 3), np.nan, dtype=np.float32)
    # For best performance, put the target chain first in concatenation.
    coords_list = [coords[target_chain_id]]
    for chain_id in coords:
        if chain_id == target_chain_id:
            continue
        coords_list.append(pad_coords)
        coords_list.append(coords[chain_id])
    coords_concatenated = np.concatenate(coords_list, axis=0)
    return coords_concatenated


def concatenate_coords_HL(
    coords_dict, seq_dict, pos_dict, posins_dict, heavy_chain_id, padding_length=10
):
    """
    Args:
        coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
            coordinates representing the backbone of each chain
        seq_dict: Dictionary mapping chain ids to native sequences of each chain
        pos_dict: Residue position IDs (e.g. IMGT numbering)
        heavy_chain_id: Heavy chain id, always start
        padding_length: Length of padding between concatenated chains
    Returns:
        Tuple (coords, seq, pos)
            - coords_concatenated is an L x 3 x 3 array for N, CA, C coordinates, a
              concatenation of the chains with padding in between
            - seq_str is the extracted sequence, with padding tokens inserted
              between the concatenated chains
            - pos_concatenated is the residue position IDs, with NaNs inserted
    """

    pad_coords = np.full((padding_length, 3, 3), np.inf, dtype=np.float32)
    pad_pos = np.full((padding_length), np.nan, dtype=np.float32)
    pad_seq = "-" * padding_length

    # Make sure always only 2 chain (H and L)
    assert len(coords_dict) == 2

    # Add heavy chain
    coords_list = [coords_dict[heavy_chain_id]]
    pos_list = [pos_dict[heavy_chain_id]]
    # Position + insertion (string) # MH
    posins_list = [posins_dict[heavy_chain_id]]
    seq_str = seq_dict[heavy_chain_id]

    # Add light chain, with 10x padding between H and L chains
    for chain_id in coords_dict:
        if chain_id == heavy_chain_id:
            continue

        coords_list.append(pad_coords)
        coords_list.append(coords_dict[chain_id])
        pos_list.append(pad_pos)
        pos_list.append(pos_dict[chain_id])
        # Position + insertion (string) # MH
        posins_list.append(pad_pos)
        posins_list.append(posins_dict[chain_id])
        seq_str += pad_seq
        seq_str += seq_dict[chain_id]

    coords_concatenated = np.concatenate(coords_list, axis=0)
    pos_concatenated = np.concatenate(pos_list, axis=0)
    # Position + insertion (string) # MH
    posins_concatenated = np.concatenate(posins_list, axis=0)

    if not len(seq_str) == len(coords_concatenated) == len(pos_concatenated):
        print(
            f"Length mismatch: seq_str {len(seq_str)} coords_concatenated {len(coords_concatenated)} pos_concatenated {len(pos_concatenated)}"
        )
        raise ValueError

    return coords_concatenated, seq_str, pos_concatenated, posins_concatenated


def concatenate_coords_any(
    coords_dict, seq_dict, pos_dict, posins_dict, chains, padding_length=10
):
    """
    Args:
        coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
            coordinates representing the backbone of each chain
        seq_dict: Dictionary mapping chain ids to native sequences of each chain
        pos_dict: Residue position IDs (e.g. IMGT numbering)
        first_chain_id: First chain id, always start
        padding_length: Length of padding between concatenated chains
    Returns:
        Tuple (coords, seq, pos)
            - coords_concatenated is an L x 3 x 3 array for N, CA, C coordinates, a
              concatenation of the chains with padding in between
            - seq_str is the extracted sequence, with padding tokens inserted
              between the concatenated chains
            - pos_concatenated is the residue position IDs, with NaNs inserted
    """

    pad_coords = np.full((padding_length, 3, 3), np.inf, dtype=np.float32)
    pad_pos = np.full((padding_length), np.nan, dtype=np.float32)
    pad_seq = "-" * padding_length

    # Add heavy chain
    first_chain_id = chains[0]
    coords_list = [coords_dict[first_chain_id]]
    pos_list = [pos_dict[first_chain_id]]
    # Position + insertion (string) # MH
    posins_list = [posins_dict[first_chain_id]]
    seq_str = seq_dict[first_chain_id]

    # Add remaining chains, with 10x padding between H and L chains
    for chain_id in chains:
        if chain_id == first_chain_id:
            continue

        coords_list.append(pad_coords)
        coords_list.append(coords_dict[chain_id])
        pos_list.append(pad_pos)
        pos_list.append(pos_dict[chain_id])
        # Position + insertion (string) # MH
        posins_list.append(pad_pos)
        posins_list.append(posins_dict[chain_id])
        seq_str += pad_seq
        seq_str += seq_dict[chain_id]

    coords_concatenated = np.concatenate(coords_list, axis=0)
    pos_concatenated = np.concatenate(pos_list, axis=0)
    # Position + insertion (string) # MH
    posins_concatenated = np.concatenate(posins_list, axis=0)

    if not len(seq_str) == len(coords_concatenated) == len(pos_concatenated):
        print(
            f"Length mismatch: seq_str {len(seq_str)} coords_concatenated {len(coords_concatenated)} pos_concatenated {len(pos_concatenated)}"
        )
        raise ValueError

    return coords_concatenated, seq_str, pos_concatenated, posins_concatenated


def sample_sequence_in_complex(
    model, coords, target_chain_id, temperature=1.0, padding_length=10
):
    """
    Samples sequence for one chain in a complex.
    Args:
        model: An instance of the GVPTransformer model
        coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
            coordinates representing the backbone of each chain
        target_chain_id: The chain id to sample sequences for
        padding_length: padding length in between chains
    Returns:
        Sampled sequence for the target chain
    """
    target_chain_len = coords[target_chain_id].shape[0]
    all_coords = _concatenate_coords(coords, target_chain_id)

    # Supply padding tokens for other chains to avoid unused sampling for speed
    padding_pattern = ["<pad>"] * all_coords.shape[0]
    for i in range(target_chain_len):
        padding_pattern[i] = "<mask>"
    sampled = model.sample(
        all_coords, partial_seq=padding_pattern, temperature=temperature
    )
    sampled = sampled[:target_chain_len]
    return sampled


def score_sequence_in_complex(
    model, alphabet, coords, target_chain_id, target_seq, padding_length=10
):
    """
    Scores sequence for one chain in a complex.
    Args:
        model: An instance of the GVPTransformer model
        alphabet: Alphabet for the model
        coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
            coordinates representing the backbone of each chain
        target_chain_id: The chain id to sample sequences for
        target_seq: Target sequence for the target chain for scoring.
        padding_length: padding length in between chains
    Returns:
        Tuple (ll_fullseq, ll_withcoord)
        - ll_fullseq: Average log-likelihood over the full target chain
        - ll_withcoord: Average log-likelihood in target chain excluding those
            residues without coordinates
    """
    all_coords = _concatenate_coords(coords, target_chain_id)

    loss, target_padding_mask = get_sequence_loss(
        model, alphabet, all_coords, target_seq
    )
    ll_fullseq = -np.sum(loss * ~target_padding_mask) / np.sum(~target_padding_mask)

    # Also calculate average when excluding masked portions
    coord_mask = np.all(np.isfinite(coords[target_chain_id]), axis=(-1, -2))
    ll_withcoord = -np.sum(loss * coord_mask) / np.sum(coord_mask)
    return ll_fullseq, ll_withcoord


def get_encoder_output_for_complex(model, alphabet, coords, target_chain_id):
    """
    Args:
        model: An instance of the GVPTransformer model
        alphabet: Alphabet for the model
        coords: Dictionary mapping chain ids to L x 3 x 3 array for N, CA, C
            coordinates representing the backbone of each chain
        target_chain_id: The chain id to sample sequences for
    Returns:
        Dictionary mapping chain id to encoder output for each chain
    """
    all_coords = _concatenate_coords(coords, target_chain_id)
    all_rep = get_encoder_output(model, alphabet, all_coords)
    target_chain_len = coords[target_chain_id].shape[0]
    return all_rep[:target_chain_len]


# MH
def get_ab_HL_coords_seq_pos_posins(pdb_path: str, Hchain: str, Lchain: str):
    """Get combined antibody H/L chain coords, seq, pos, posins from pdb_path"""

    coords_dict, seq_dict, pos_dict, posins_dict = load_complex_coords(
        pdb_path, [Hchain, Lchain]
    )
    (
        coords_concatenated,
        seq_concatenated,
        pos_concatenated,
        posins_concatenated,
    ) = concatenate_coords_HL(
        coords_dict, seq_dict, pos_dict, posins_dict, heavy_chain_id=Hchain
    )

    return coords_concatenated, seq_concatenated, pos_concatenated, posins_concatenated



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
import urllib
import warnings
from argparse import Namespace
from pathlib import Path

import torch

import antifold.esm


def _has_regression_weights(model_name):
    """Return whether we expect / require regression weights;
    Right now that is all models except ESM-1v, ESM-IF, and partially trained ESM2 models
    """
    return not (
        "esm1v" in model_name
        or "esm_if" in model_name
        or "270K" in model_name
        or "500K" in model_name
    )


def load_model_and_alphabet(model_name):
    if model_name.endswith(".pt"):  # treat as filepath
        return load_model_and_alphabet_local(model_name)
    else:
        return load_model_and_alphabet_hub(model_name)


def load_hub_workaround(url):
    try:
        data = torch.hub.load_state_dict_from_url(
            url, progress=False, map_location="cpu"
        )
    except RuntimeError:
        # Pytorch version issue - see https://github.com/pytorch/pytorch/issues/43106
        fn = Path(url).name
        data = torch.load(
            f"{torch.hub.get_dir()}/checkpoints/{fn}",
            map_location="cpu",
        )
    except urllib.error.HTTPError as e:
        raise Exception(
            f"Could not load {url}, check if you specified a correct model name?"
        )
    return data


def load_regression_hub(model_name):
    url = f"https://dl.fbaipublicfiles.com/fair-esm/regression/{model_name}-contact-regression.pt"
    regression_data = load_hub_workaround(url)
    return regression_data


def _download_model_and_regression_data(model_name):
    url = f"https://dl.fbaipublicfiles.com/fair-esm/models/{model_name}.pt"
    model_data = load_hub_workaround(url)
    if _has_regression_weights(model_name):
        regression_data = load_regression_hub(model_name)
    else:
        regression_data = None
    return model_data, regression_data


def load_model_and_alphabet_hub(model_name):
    model_data, regression_data = _download_model_and_regression_data(model_name)
    return load_model_and_alphabet_core(model_name, model_data, regression_data)


def load_model_and_alphabet_local(model_location):
    """Load from local path. The regression weights need to be co-located"""
    model_location = Path(model_location)
    model_data = torch.load(str(model_location), map_location="cpu")
    model_name = model_location.stem
    if _has_regression_weights(model_name):
        regression_location = (
            str(model_location.with_suffix("")) + "-contact-regression.pt"
        )
        regression_data = torch.load(regression_location, map_location="cpu")
    else:
        regression_data = None
    return load_model_and_alphabet_core(model_name, model_data, regression_data)


def has_emb_layer_norm_before(model_state):
    """Determine whether layer norm needs to be applied before the encoder"""
    return any(
        k.startswith("emb_layer_norm_before") for k, param in model_state.items()
    )


def _load_model_and_alphabet_core_v1(model_data):
    import antifold.esm  # since esm.inverse_folding is imported below, you actually have to re-import esm here

    alphabet = antifold.esm.Alphabet.from_architecture(model_data["args"].arch)

    if "invariant_gvp" in model_data["args"].arch:
        import antifold.esm.inverse_folding

        model_type = antifold.esm.inverse_folding.gvp_transformer.GVPTransformerModel
        model_args = vars(model_data["args"])  # convert Namespace -> dict

        def update_name(s):
            # Map the module names in checkpoints trained with internal code to
            # the updated module names in open source code
            s = s.replace("W_v", "embed_graph.embed_node")
            s = s.replace("W_e", "embed_graph.embed_edge")
            s = s.replace("embed_scores.0", "embed_confidence")
            s = s.replace("embed_score.", "embed_graph.embed_confidence.")
            s = s.replace("seq_logits_projection.", "")
            s = s.replace("embed_ingraham_features", "embed_dihedrals")
            s = s.replace("embed_gvp_in_local_frame.0", "embed_gvp_output")
            s = s.replace("embed_features_in_local_frame.0", "embed_gvp_input_features")
            return s

        model_state = {
            update_name(sname): svalue
            for sname, svalue in model_data["model"].items()
            if "version" not in sname
        }

    else:
        raise ValueError("Unknown architecture selected")

    model = model_type(
        Namespace(**model_args),
        alphabet,
    )

    return model, alphabet, model_state


def _load_IF1_local():
    import antifold.esm  # since esm.inverse_folding is imported below, you actually have to re-import esm here

    alphabet = antifold.esm.Alphabet.from_architecture("vt_medium_with_invariant_gvp")

    import antifold.esm.inverse_folding

    model_type = antifold.esm.inverse_folding.gvp_transformer.GVPTransformerModel
    model_args = IF1_dict  # convert Namespace -> dict

    def update_name(s):
        # Map the module names in checkpoints trained with internal code to
        # the updated module names in open source code
        s = s.replace("W_v", "embed_graph.embed_node")
        s = s.replace("W_e", "embed_graph.embed_edge")
        s = s.replace("embed_scores.0", "embed_confidence")
        s = s.replace("embed_score.", "embed_graph.embed_confidence.")
        s = s.replace("seq_logits_projection.", "")
        s = s.replace("embed_ingraham_features", "embed_dihedrals")
        s = s.replace("embed_gvp_in_local_frame.0", "embed_gvp_output")
        s = s.replace("embed_features_in_local_frame.0", "embed_gvp_input_features")
        return s

    model = model_type(
        Namespace(**model_args),
        alphabet,
    )

    return model, alphabet


def load_model_and_alphabet_core(model_name, model_data, regression_data=None):
    if regression_data is not None:
        model_data["model"].update(regression_data["model"])

    model, alphabet, model_state = _load_model_and_alphabet_core_v1(model_data)

    expected_keys = set(model.state_dict().keys())
    found_keys = set(model_state.keys())

    if regression_data is None:
        expected_missing = {
            "contact_head.regression.weight",
            "contact_head.regression.bias",
        }
        error_msgs = []
        missing = (expected_keys - found_keys) - expected_missing
        if missing:
            error_msgs.append(f"Missing key(s) in state_dict: {missing}.")
        unexpected = found_keys - expected_keys
        if unexpected:
            error_msgs.append(f"Unexpected key(s) in state_dict: {unexpected}.")

        if error_msgs:
            raise RuntimeError(
                "Error(s) in loading state_dict for {}:\n\t{}".format(
                    model.__class__.__name__, "\n\t".join(error_msgs)
                )
            )
        if expected_missing - found_keys:
            warnings.warn(
                "Regression weights not found, predicting contacts will not produce correct results."
            )

    model.load_state_dict(model_state, strict=regression_data is not None)

    return model, alphabet


def esm_if1_gvp4_t16_142M_UR50():
    """Inverse folding model with 142M params, with 4 GVP-GNN layers, 8
    Transformer encoder layers, and 8 Transformer decoder layers, trained on
    CATH structures and 12 million alphafold2 predicted structures from UniRef50
    sequences.

    Returns a tuple of (Model, Alphabet).
    """
    return load_model_and_alphabet_hub("esm_if1_gvp4_t16_142M_UR50")


IF1_dict = {
'no_progress_bar': False,
'log_interval': 100,
'log_format': 'json',
'tensorboard_logdir': '',
'seed': 1238123,
'cpu': False,
'tpu': False,
'bf16': False,
'fp16': False,
'memory_efficient_bf16': False,
'memory_efficient_fp16': False,
'fp16_no_flatten_grads': False,
'fp16_init_scale': 128,
'fp16_scale_window': None,
'fp16_scale_tolerance': 0.0,
'min_loss_scale': 0.0001,
'threshold_loss_scale': None,
'user_dir': None,
'empty_cache_freq': 0,
'all_gather_list_size': 16384,
'model_parallel_size': 1,
'checkpoint_suffix': '',
'quantization_config_path': None,
'profile': False,
'criterion': 'inverse_folding_loss',
'tokenizer': None,
'bpe': None,
'optimizer': 'adam',
'lr_scheduler': 'inverse_sqrt',
'scoring': 'bleu',
'task': 'inverse_folding',
'num_workers': 8,
'skip_invalid_size_inputs_valid_test': False,
'max_tokens': 4096,
'max_sentences': None,
'required_batch_size_multiple': 8,
'dataset_impl': None,
'data_buffer_size': 10,
'train_subset': 'cath-chains-maxlen500/train,alphafold_relaxed_12M_flat/train',
'valid_subset': 'cath-chains-maxlen500/valid,cath-chains-maxlen500/short-valid,cath-chains-maxlen500/single-valid,alphafold_relaxed_12M_flat/valid',
'validate_interval': 1,
'validate_interval_updates': 0,
'validate_after_updates': 0,
'fixed_validation_seed': None,
'disable_validation': False,
'max_tokens_valid': 4096,
'max_sentences_valid': None,
'curriculum': 0,
'distributed_world_size': 32,
'distributed_rank': 0,
'distributed_backend': 'nccl',
'distributed_init_method': 'tcp://learnfair7689:14349',
'distributed_port': 14349,
'device_id': 0,
'distributed_no_spawn': False,
'ddp_backend': 'c10d',
'bucket_cap_mb': 25,
'fix_batches_to_gpus': False,
'find_unused_parameters': False,
'fast_stat_sync': False,
'broadcast_buffers': False,
'heartbeat_timeout': -1,
'distributed_wrapper': 'DDP',
'slowmo_momentum': None,
'slowmo_algorithm': 'LocalSGD',
'localsgd_frequency': 3,
'nprocs_per_node': 8,
'zero_sharding': 'none',
'arch': 'vt_medium_with_invariant_gvp',
'max_epoch': 200,
'max_update': 0,
'stop_time_hours': 0,
'clip_norm': 0.0,
'sentence_avg': False,
'update_freq': [1],
'lr': [0.001],
'min_lr': 1e-09,
'use_bmuf': False,
'save_dir': 'ablation3/ablation_transformer.seed0.s0.9.t0.0.cath+alphafold.ratios0.135.sub1.000.afns0.1.medium_with_invariant_gvp.bs4096.uf1.lr0.001.sqrtlr.warmup5000.mp0.15.geop0.05.upperspan30.ngpu32',
'restore_file': 'checkpoint_last.pt',
'finetune_from_model': None,
'reset_dataloader': False,
'reset_lr_scheduler': False,
'reset_meters': False,
'reset_optimizer': False,
'optimizer_overrides': '{}',
'save_interval': 1,
'save_interval_updates': 0,
'keep_interval_updates': -1,
'keep_last_epochs': 5,
'keep_best_checkpoints': -1,
'no_save': False,
'no_epoch_checkpoints': False,
'no_last_checkpoints': False,
'no_save_optimizer_state': False,
'best_checkpoint_metric': 'loss',
'maximize_best_checkpoint_metric': False,
'patience': -1,
'no_mid_epoch_validate': False,
'no_token_positional_embeddings': False,
'no_cross_attention': False,
'cross_self_attention': False,
'encoder_layerdrop': 0,
'decoder_layerdrop': 0,
'encoder_layers_to_keep': None,
'decoder_layers_to_keep': None,
'quant_noise_pq': 0,
'quant_noise_pq_block_size': 8,
'quant_noise_scalar': 0,
'joint_generation': False,
'et_gvp_as_feedforward': False,
'rescale_opt': 'token',
'use_inf_norm': False,
'adam_betas': '[0.9,0.999]',
'adam_eps': 1e-08,
'weight_decay': 0.0,
'use_old_adam': False,
'warmup_updates': 5000,
'warmup_init_lr': 1e-07,
'data': '/large_experiments/protein/data/inverse-folding/splits',
'binning_scheme': 'trRosetta_v2',
'include_angle': True,
'from_3d': True,
'in_memory_dataset': False,
'include_torsion': True,
'prepend_token': True,
'max_positions': 1024,
'train_sample_ratios': '1,0.135',
'valid_sample_ratios': '1,1,1,0.135',
'subsample_train': '1,1.000',
'train_crop_size': -1,
'min_num_source_residues': 0,
'partial_seq_masking': False,
'predict_next_only': False,
'test_time_coord_masking': False,
'source_coord_mask_prob': 0.15,
'source_coord_span_masking_span_upper': 30,
'source_coord_span_masking_span_lower': 1,
'source_coord_span_masking_geometric_p': 0.05,
'train_input_noises': '0,0.1',
'random_rotation': True,
'source_confidence_threshold': 0.9,
'target_confidence_threshold': 0.0,
'n_cond_score_buckets': 22,
'gvp_conditioning_encoder': True,
'gvp_conditioning_score_num_rbf': 16,
'gvp_node_input_dim_scalar': 7,
'gvp_node_input_dim_vector': 3,
'gvp_edge_input_dim_scalar': 34,
'gvp_edge_input_dim_vector': 1,
'gvp_vector_gate': True,
'gvp_attention_heads': 0,
'gvp_n_message_gvps': 3,
'gvp_n_edge_gvps_first_layer': 0,
'gvp_n_edge_gvps': 0,
'gvp_eps': 0.0001,
'gvp_layernorm': True,
'gvp_no_edge_orientation': False,
'gvp_conv_no_scalar_activation': False,
'gvp_conv_no_vector_activation': False,
'gvp_ignore_edges_without_coords': True,
'no_seed_provided': False,
'gvp_node_hidden_dim_scalar': 1024,
'gvp_node_hidden_dim_vector': 256,
'gvp_num_encoder_layers': 4,
'gvp_dropout': 0.1,
'gvp_top_k_neighbors': 30,
'gvp_edge_hidden_dim_scalar': 32,
'gvp_edge_hidden_dim_vector': 1,
'embed_gvp_in_global_frame': False,
'embed_gvp_in_local_frame': True,
'embed_ingraham_features': True,
'embed_rotation_frames': False,
'embed_rotation_quaternions': False,
'embed_features_in_global_frame': False,
'embed_features_in_local_frame': True,
'embed_scores': True,
'encoder_embed_dim': 512,
'encoder_ffn_embed_dim': 2048,
'encoder_attention_heads': 8,
'encoder_layers': 8,
'decoder_embed_dim': 512,
'decoder_ffn_embed_dim': 2048,
'decoder_attention_heads': 8,
'decoder_layers': 8,
'attention_dropout': 0.1,
'encoder_normalize_before': True,
'decoder_normalize_before': True,
'pretrained_gvp_stem': None,
'frozen_gvp_stem': False,
'pretrained_decoder': None,
'frozen_decoder': False,
'encoder_edge_attn_layers': 0,
'edge_attn_distance_cutoff': -1,
'edge_embed_dim': 0,
'embed_edge_vectors': False,
'distance_noise': 0.0,
'gvp_distance_noise': 0.0,
'equivariant_attn_layers': 0,
'embed_patch_layers': 0,
'encoder_embed_path': None,
'encoder_learned_pos': False,
'decoder_embed_path': None,
'decoder_learned_pos': False,
'activation_dropout': 0.0,
'activation_fn': 'relu',
'dropout': 0.1,
'adaptive_softmax_cutoff': None,
'adaptive_softmax_dropout': 0,
'share_decoder_input_output_embed': False,
'share_all_embeddings': False,
'adaptive_input': False,
'decoder_output_dim': 512,
'decoder_input_dim': 512,
'no_scale_embedding': False,
'layernorm_embedding': False,
'tie_adaptive_weights': False,
'max_source_positions': 1024,
'max_target_positions': 1024
}


import logging
import os
import sys
# import warnings
import urllib.request
from pathlib import Path

ROOT_PATH = Path(os.path.dirname(__file__)).parent
sys.path.insert(0, ROOT_PATH)

from argparse import ArgumentParser, RawTextHelpFormatter

import numpy as np
import pandas as pd

from antifold.antiscripts import (df_logits_to_logprobs,
                                  extract_chains_biotite, generate_pdbs_csv,
                                  get_pdbs_logits, load_model,
                                  sample_from_df_logits, write_fasta_to_dir,
                                  visualize_mutations)

log = logging.getLogger(__name__)


def cmdline_args():
    # Make parser object
    usage = f"""
# Run AntiFold on single PDB (or CIF) file
python antifold/main.py \
    --out_dir output/single_pdb \
    --pdb_file data/pdbs/6y1l_imgt.pdb \
    --heavy_chain H \
    --light_chain L

# Run AntiFold on an antibody-antigen complex (enables custom_chain_mode)
python antifold/main.py \
    --out_dir output/antibody_antigen \
    --pdb_file data/antibody_antigen/3hfm.pdb \
    --heavy_chain H \
    --light_chain L \
    --antigen_chain Y

# Run AntiFold on a folder of PDB/CIFs (specify chains to run in CSV file)
# and consider extra antigen chains
python antifold/main.py \
    --out_dir output/antibody_antigen \
    --pdbs_csv data/antibody_antigen.csv \
    --pdb_dir data/antibody_antigen \
    --custom_chain_mode
    """
    p = ArgumentParser(
        description="Predict antibody variable domain inverse folding probabilities and sample sequences with maintained fold.\nPDB structures should be IMGT-numbered, paired heavy and light chain variable domains (positions 1-128).\n\nFor IMGT numbering PDBs use SAbDab or https://opig.stats.ox.ac.uk/webapps/sabdab-sabpred/sabpred/anarci/",
        formatter_class=RawTextHelpFormatter,
        usage=usage,
    )

    def is_valid_path(parser, arg):
        if not os.path.exists(arg):
            parser.error(f"Path {arg} does not exist!")
        else:
            return arg

    def is_valid_dir(parser, arg):
        if not os.path.isdir(arg):
            parser.error(f"Directory {arg} does not exist!")
        else:
            return arg

    p.add_argument(
        "--pdb_file",
        help="Input PDB file (for single PDB predictions)",
        type=lambda x: is_valid_path(p, x),
    )

    p.add_argument(
        "--nanobody_chain", help="Antibody nanobody chain (for single PDB predictions)",
    )

    p.add_argument(
        "--heavy_chain", help="Antibody heavy chain (for single PDB predictions)",
    )

    p.add_argument(
        "--light_chain", help="Antibody light chain (for single PDB predictions)",
    )

    p.add_argument(
        "--antigen_chain", help="Antigen chain (optional)",
    )

    p.add_argument(
        "--pdbs_csv",
        help="Input CSV file with PDB names and H/L chains (multi-PDB predictions)",
        type=lambda x: is_valid_path(p, x),
    )

    p.add_argument(
        "--pdb_dir",
        help="Directory with input PDB files (multi-PDB predictions)",
        type=lambda x: is_valid_dir(p, x),
    )

    p.add_argument(
        "--out_dir", default="antifold_output", help="Output directory",
    )

    p.add_argument(
        "--regions",
        default="CDR1 CDR2 CDR3",
        help="Space-separated regions to mutate. Default 'CDR1 CDR2 CDR3H'",
    )

    p.add_argument(
        "--num_seq_per_target",
        default=0,
        type=int,
        help="Number of sequences to sample from each antibody PDB (default 0)",
    )

    p.add_argument(
        "--sampling_temp",
        default="0.20",
        help="A string of temperatures e.g. '0.20 0.25 0.50' (default 0.20). Sampling temperature for amino acids. Suggested values 0.10, 0.15, 0.20, 0.25, 0.30. Higher values will lead to more diversity.",
    )

    p.add_argument(
        "--limit_variation",
        default=False,
        action="store_true",
        help="Limit variation to as many mutations as expected from temperature sampling",
    )

    p.add_argument(
        "--extract_embeddings",
        default=False,
        action="store_true",
        help="Extract per-residue embeddings from AntiFold / ESM-IF1",
    )

    p.add_argument(
        "--custom_chain_mode",
        default=False,
        action="store_true",
        help="Run all specified chains (for antibody-antigen complexes or any combination of chains)",
    )

    p.add_argument(
        "--exclude_heavy", action="store_true", help="Exclude heavy chain from sampling"
    )

    p.add_argument(
        "--exclude_light", action="store_true", help="Exclude light chain from sampling"
    )

    p.add_argument(
        "--batch_size", default=1, type=int, help="Batch-size to use",
    )

    p.add_argument(
        "--num_threads",
        default=0,
        type=int,
        help="Number of CPU threads to use for parallel processing (0 = all available)",
    )

    p.add_argument(
        "--seed", default=42, type=int, help="Seed for reproducibility",
    )

    p.add_argument(
        "--model_path",
        default="",
        help="Alternative model weights (default models/model.pt). See --use_esm_if1_weights flag to use ESM-IF1 weights instead of AntiFold",
    )

    p.add_argument(
        "--esm_if1_mode",
        default=False,
        action="store_true",
        help="Use ESM-IF1 weights instead of AntiFold",
    )

    p.add_argument(
        "--verbose", default=1, type=int, help="Verbose printing",
    )

    return p.parse_args()


def sample_pdbs(
    model,
    pdbs_csv_or_dataframe,
    regions_to_mutate,
    pdb_dir="data/pdbs",
    out_dir="output/sampled",
    sample_n=10,
    sampling_temp=0.50,
    limit_expected_variation=False,
    exclude_heavy=False,
    exclude_light=False,
    batch_size=1,
    extract_embeddings=False,
    custom_chain_mode=False,
    num_threads=0,
    seed=42,
    save_flag=False,
):
    # Predict with CSV on folder of solved (SAbDab) structures
    df_logits_list = get_pdbs_logits(
        model=model,
        pdbs_csv_or_dataframe=pdbs_csv_or_dataframe,
        pdb_dir=pdb_dir,
        out_dir=out_dir,
        save_flag=save_flag,
        batch_size=batch_size,
        extract_embeddings=extract_embeddings,
        custom_chain_mode=custom_chain_mode,
        seed=42,
        num_threads=num_threads,
    )

    if sample_n >= 1:
        # Sample from output probabilities
        pdb_output_dict = {}
        for df_logits in df_logits_list:
            # Sample 10 sequences with a temperature of 0.50
            fasta_dict = sample_from_df_logits(
                df_logits,
                sample_n=sample_n,
                sampling_temp=sampling_temp,
                regions_to_mutate=regions_to_mutate,
                limit_expected_variation=False,
                verbose=True,
            )

            pdb_output_dict[df_logits.name] = {
                "sequences": fasta_dict,
                "logits": df_logits,
                "logprobs": df_logits_to_logprobs(df_logits),
            }

            # Write to file
            if save_flag:
                write_fasta_to_dir(fasta_dict, df_logits, out_dir=out_dir)

        return pdb_output_dict


def check_valid_input(args):
    """Checks for valid arguments"""

    # Check valid input files input arguments
    # Check either: PDB file, PDB dir or PDBs CSV inputted
    if not (args.pdb_file or args.pdb_dir):
        log.error(
            f"""Please choose one of:
        1) PDB file (--pdb_file). We heavily recommend specifying --heavy_chain [letter] and --light_chain [letter]
        2) PDB directory (--pdb_dir) and CSV file (--pdbs_csv) with columns for PDB names (pdb), H (Hchain) and L (Lchain) chains
        3) PDB directory (--pdb_dir). Warning: Will assume 1st chain is heavy, 2nd chain is light
        """
        )
        sys.exit(1)

    # # Check that AntiFold weights are downloaded
    # root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    # filename = "models/model.pt"
    # model_path = f"{root_dir}/{filename}"
    # if not os.path.exists(model_path):
    #     log.warning(
    #         f"Downloading AntiFold model weights from https://opig.stats.ox.ac.uk/data/downloads/AntiFold/models/model.pt to {model_path}"
    #     )
    #     url = "https://opig.stats.ox.ac.uk/data/downloads/AntiFold/models/model.pt"

    #     os.makedirs(f"{root_dir}/models", exist_ok=True)
    #     urllib.request.urlretrieve(url, filename)

    # Option 1: PDB file, check heavy and light chain
    if args.pdb_file:
        if not args.heavy_chain and not args.nanobody_chain:
            _pdb = os.path.splitext(os.path.basename(args.pdb_file))[0]
            log.warning(
                f"WARNING: Heavy/light chain(s) not specified for {_pdb}. Assuming 1st chain heavy, 2nd chain light."
            )
            log.warning(
                f"WARNING: Specify manually with e.g. --heavy_chain H --light_chain L"
            )

    # Option 2: Check PDBs in PDB dir and CSV formatted correctly
    elif args.pdb_dir and args.pdbs_csv:

        # Run all chains specified in the CSV file
        args.custom_chain_mode = True

        # Check CSV formatting
        df = pd.read_csv(args.pdbs_csv, comment="#")
        if (
            not df.columns.isin(["pdb", "Hchain", "Lchain"]).sum() >= 3
            and not args.custom_chain_mode
        ):
            log.error(
                f"Multi-PDB input: Please specify CSV  with columns ['pdb', 'Hchain', 'Lchain'] with PDB names (no extension), H and L chains"
            )
            log.error(f"CSV columns: {df.columns}")
            sys.exit(1)

        # Check PDBs exist
        missing = 0
        for i, _pdb in enumerate(df["pdb"].values):
            pdb_path = f"{args.pdb_dir}/{_pdb}.pdb"

            # Check for PDB/CIF file
            pdb_path = (
                pdb_path if os.path.exists(pdb_path) else f"{args.pdb_dir}/{_pdb}.cif"
            )

            if not os.path.exists(pdb_path):
                log.warning(
                    f"WARNING: Unable to find PDB/CIF file ({missing+1}): {pdb_path}"
                )
                missing += 1

        if missing >= 1:
            log.error(
                f"WARNING: Missing {missing} PDB/CIFs specified in {args.pdbs_csv} but not found in {args.pdb_dir}"
            )
            sys.exit(1)

    # Option 3: PDB directory only, infer chains
    elif args.pdb_dir:
        _dir = os.path.dirname(args.pdb_dir)
        log.warning(
            f"WARNING: Heavy/light chains not specified for PDB/CIF files in folder {_dir}. Assuming 1st chain heavy, 2nd chain light."
        )
        log.warning(f"WARNING: Specify manually with --pdbs_csv CSV file")

    # ESM-IF1 mode
    if args.esm_if1_mode:
        args.model_path = "ESM-IF1"
        args.custom_chain_mode = True

        if args.out_dir == "antifold_output":
            args.out_dir = "esmif1_output"

        log.info(
            f"NOTE: ESM-IF1 mode enabled, will use ESM-IF1 weights and run all specified chains"
        )


def main(args):
    """Predicts antibody heavy and light chain inverse folding probabilities"""

    # Create output directory
    os.makedirs(args.out_dir, exist_ok=True)

    # Try reading in regions
    regions_to_mutate = []
    for region in args.regions.split(" "):
        # Either interpret as positions (ints)
        try:
            regions_to_mutate.append(int(region))
        # Or as regions (strings)
        except ValueError:
            regions_to_mutate.append(region)

    # Try reading in sampling temperatures
    try:
        args.sampling_temp = [float(t) for t in args.sampling_temp.split(" ")]
    except ValueError:
        raise Exception(
            "Sampling temperature must be a float or space-separated floats, e.g. '0.20 0.25 0.50'"
        )

    # No chains provided: assume 1st chain is heavy, 2nd is light
    # Option 1: Single PDB
    if args.pdb_file:

        _pdb = os.path.splitext(os.path.basename(args.pdb_file))[0]

        # Nanobody requires custom_chain_mode
        if args.nanobody_chain:
            args.custom_chain_mode = True
            args.heavy_chain = args.nanobody_chain

        # No chains specified, assume 1st heavy, 2nd light (unless single-chain mode)
        elif not args.heavy_chain:
            args.heavy_chain, args.light_chain = extract_chains_biotite(args.pdb_file)[
                :2
            ]
            log.warning(
                f"{_pdb}: assuming heavy_chain {args.heavy_chain}, light_chain {args.light_chain}"
            )

        pdbs_csv = pd.DataFrame(
            {"pdb": _pdb, "Hchain": args.heavy_chain, "Lchain": args.light_chain,},
            index=[0],
        )

        # Single heavy chain requires custom_chain_mode
        if args.heavy_chain and not args.light_chain:
            args.custom_chain_mode = True

        # Antigen chain requires custom_chain_mode
        if args.antigen_chain:
            pdbs_csv.loc[0, "Agchain"] = args.antigen_chain
            args.custom_chain_mode = True

        # PDB dir is the directory of the PDB file
        pdb_dir = os.path.dirname(args.pdb_file)

    # Option 2: PDB dir and CSV file
    elif args.pdb_dir and args.pdbs_csv:
        pdbs_csv = pd.read_csv(args.pdbs_csv, comment="#")
        pdb_dir = args.pdb_dir

    # Option 3: PDB dir and no CSV file (infer chains)
    else:
        pdb_dir = args.pdb_dir

        # custom_chain_mode consider all (10) chains in file
        if args.custom_chain_mode:
            pdbs_csv = generate_pdbs_csv(args.pdb_dir, max_chains=10)

        # Other only consider 1st (heavy) chain and 2nd (light) chain
        else:
            pdbs_csv = generate_pdbs_csv(args.pdb_dir, max_chains=2)

    # Extra: Sample sequences with num_seq_per_target >= 1
    if args.num_seq_per_target >= 1:
        log.info(
            f"Will sample {args.num_seq_per_target} sequences from {len(pdbs_csv.values)} PDBs at temperature(s) {args.sampling_temp} and regions: {regions_to_mutate}"
        )

    # Load AntiFold or ESM-IF1 model
    # Infer model from file path
    model = load_model()

    # Get dict with PDBs, sampled sequences and logits / log_odds DataFrame
    pdb_output_dict = sample_pdbs(
        model=model,
        pdbs_csv_or_dataframe=pdbs_csv,
        pdb_dir=pdb_dir,
        regions_to_mutate=regions_to_mutate,
        out_dir=args.out_dir,
        sample_n=args.num_seq_per_target,
        sampling_temp=args.sampling_temp,
        limit_expected_variation=args.limit_variation,
        exclude_heavy=args.exclude_heavy,
        exclude_light=args.exclude_light,
        batch_size=args.batch_size,
        extract_embeddings=args.extract_embeddings,
        custom_chain_mode=args.custom_chain_mode,
        num_threads=args.num_threads,
        seed=args.seed,
        save_flag=True,
    )


if __name__ == "__main__":
    args = cmdline_args()

    # Log to file and stdout
    os.makedirs(args.out_dir, exist_ok=True)
    log_path = os.path.abspath(f"{args.out_dir}/log.txt")

    logging.basicConfig(
        level=logging.INFO,
        format="[{asctime}] {message}",
        style="{",
        handlers=[
            logging.FileHandler(filename=log_path, mode="w"),
            logging.StreamHandler(stream=sys.stdout),
        ],
    )
    log = logging.getLogger(__name__)

    # INFO prints total summary and errors (default)
    if args.verbose == 1:
        logging.getLogger().setLevel(logging.INFO)

    # DEBUG prints every major step
    elif args.verbose >= 2:
        logging.getLogger().setLevel(logging.DEBUG)

    # Check valid input
    try:
        log.info(f"Running inverse folding on PDB/CIFs ...")
        check_valid_input(args)
        main(args)

    except Exception as E:
        log.exception(f"Prediction encountered an unexpected error: {E}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import math
from typing import List, Sequence, Tuple

import biotite.structure
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data as data
from biotite.sequence import ProteinSequence
from biotite.structure import filter_backbone, get_chains
from biotite.structure.io import pdb, pdbx
from biotite.structure.residues import get_residues
from scipy.spatial import transform
from scipy.stats import special_ortho_group

from antifold.esm.data import BatchConverter


def load_structure(fpath, chain=None):
    """
    Args:
        fpath: filepath to either pdb or cif file
        chain: the chain id or list of chain ids to load
    Returns:
        biotite.structure.AtomArray
    """
    if fpath.endswith("cif"):
        with open(fpath) as fin:
            pdbxf = pdbx.PDBxFile.read(fin)
        structure = pdbx.get_structure(pdbxf, model=1)
    elif fpath.endswith("pdb"):
        with open(fpath) as fin:
            pdbf = pdb.PDBFile.read(fin)
        structure = pdb.get_structure(pdbf, model=1)
    bbmask = filter_backbone(structure)
    structure = structure[bbmask]
    all_chains = get_chains(structure)
    if len(all_chains) == 0:
        raise ValueError("No chains found in the input file.")
    if chain is None:
        chain_ids = all_chains
    elif isinstance(chain, list):
        chain_ids = chain
    else:
        chain_ids = [chain]
    for chain in chain_ids:
        if chain not in all_chains:
            raise ValueError(f"Chain {chain} not found in input file")
    chain_filter = [a.chain_id in chain_ids for a in structure]
    structure = structure[chain_filter]
    return structure


def extract_coords_from_structure(structure: biotite.structure.AtomArray):
    """
    Args:
        structure: An instance of biotite AtomArray
    Returns:
        Tuple (coords, seq)
            - coords is an L x 3 x 3 array for N, CA, C coordinates
            - seq is the extracted sequence
    """
    coords = get_atom_coords_residuewise(["N", "CA", "C"], structure)
    residue_identities = get_residues(structure)[1]
    seq = "".join([ProteinSequence.convert_letter_3to1(r) for r in residue_identities])
    return coords, seq


def load_coords(fpath, chain):
    """
    Args:
        fpath: filepath to either pdb or cif file
        chain: the chain id
    Returns:
        Tuple (coords, seq)
            - coords is an L x 3 x 3 array for N, CA, C coordinates
            - seq is the extracted sequence
    """
    structure = load_structure(fpath, chain)
    return extract_coords_from_structure(structure)


def get_atom_coords_residuewise(atoms: List[str], struct: biotite.structure.AtomArray):
    """
    Example for atoms argument: ["N", "CA", "C"]
    """

    def filterfn(s, axis=None):
        filters = np.stack([s.atom_name == name for name in atoms], axis=1)
        sum = filters.sum(0)
        if not np.all(sum <= np.ones(filters.shape[1])):
            raise RuntimeError("structure has multiple atoms with same name")
        index = filters.argmax(0)
        coords = s[index].coord
        coords[sum == 0] = float("nan")
        return coords

    return biotite.structure.apply_residue_wise(struct, struct, filterfn)


def get_sequence_loss(model, alphabet, coords, seq):
    device = next(model.parameters()).device
    batch_converter = CoordBatchConverter(alphabet)
    batch = [(coords, None, seq)]
    coords, confidence, strs, tokens, padding_mask = batch_converter(
        batch, device=device
    )

    prev_output_tokens = tokens[:, :-1].to(device)
    target = tokens[:, 1:]
    target_padding_mask = target == alphabet.padding_idx
    logits, _ = model.forward(coords, padding_mask, confidence, prev_output_tokens)
    loss = F.cross_entropy(logits, target, reduction="none")
    loss = loss[0].cpu().detach().numpy()
    target_padding_mask = target_padding_mask[0].cpu().numpy()
    return loss, target_padding_mask


def score_sequence(model, alphabet, coords, seq):
    loss, target_padding_mask = get_sequence_loss(model, alphabet, coords, seq)
    ll_fullseq = -np.sum(loss * ~target_padding_mask) / np.sum(~target_padding_mask)
    # Also calculate average when excluding masked portions
    coord_mask = np.all(np.isfinite(coords), axis=(-1, -2))
    ll_withcoord = -np.sum(loss * coord_mask) / np.sum(coord_mask)
    return ll_fullseq, ll_withcoord


def get_encoder_output(model, alphabet, coords):
    device = next(model.parameters()).device
    batch_converter = CoordBatchConverter(alphabet)
    batch = [(coords, None, None)]
    coords, confidence, strs, tokens, padding_mask = batch_converter(
        batch, device=device
    )
    encoder_out = model.encoder.forward(
        coords, padding_mask, confidence, return_all_hiddens=False
    )
    # remove beginning and end (bos and eos tokens)
    return encoder_out["encoder_out"][0][1:-1, 0]


def rotate(v, R):
    """
    Rotates a vector by a rotation matrix.

    Args:
        v: 3D vector, tensor of shape (length x batch_size x channels x 3)
        R: rotation matrix, tensor of shape (length x batch_size x 3 x 3)

    Returns:
        Rotated version of v by rotation matrix R.
    """
    R = R.unsqueeze(-3)
    v = v.unsqueeze(-1)
    return torch.sum(v * R, dim=-2)


def get_rotation_frames(coords):
    """
    Returns a local rotation frame defined by N, CA, C positions.

    Args:
        coords: coordinates, tensor of shape (batch_size x length x 3 x 3)
        where the third dimension is in order of N, CA, C

    Returns:
        Local relative rotation frames in shape (batch_size x length x 3 x 3)
    """
    v1 = coords[:, :, 2] - coords[:, :, 1]
    v2 = coords[:, :, 0] - coords[:, :, 1]
    e1 = normalize(v1, dim=-1)
    u2 = v2 - e1 * torch.sum(e1 * v2, dim=-1, keepdim=True)
    e2 = normalize(u2, dim=-1)
    e3 = torch.cross(e1, e2, dim=-1)
    R = torch.stack([e1, e2, e3], dim=-2)
    return R


def nan_to_num(ts, val=0.0):
    """
    Replaces nans in tensor with a fixed value.
    """
    val = torch.tensor(val, dtype=ts.dtype, device=ts.device)
    return torch.where(~torch.isfinite(ts), val, ts)


def rbf(values, v_min, v_max, n_bins=16):
    """
    Returns RBF encodings in a new dimension at the end.
    """
    rbf_centers = torch.linspace(v_min, v_max, n_bins, device=values.device)
    rbf_centers = rbf_centers.view([1] * len(values.shape) + [-1])
    rbf_std = (v_max - v_min) / n_bins
    v_expand = torch.unsqueeze(values, -1)
    z = (values.unsqueeze(-1) - rbf_centers) / rbf_std
    return torch.exp(-(z**2))


def norm(tensor, dim, eps=1e-8, keepdim=False):
    """
    Returns L2 norm along a dimension.
    """
    return torch.sqrt(torch.sum(torch.square(tensor), dim=dim, keepdim=keepdim) + eps)


def normalize(tensor, dim=-1):
    """
    Normalizes a tensor along a dimension after removing nans.
    """
    return nan_to_num(torch.div(tensor, norm(tensor, dim=dim, keepdim=True)))


class CoordBatchConverter(BatchConverter):
    def __call__(self, raw_batch: Sequence[Tuple[Sequence, str]], device=None):
        """
        Args:
            raw_batch: List of tuples (coords, confidence, seq)
            In each tuple,
                coords: list of floats, shape L x 3 x 3
                confidence: list of floats, shape L; or scalar float; or None
                seq: string of length L
        Returns:
            coords: Tensor of shape batch_size x L x 3 x 3
            confidence: Tensor of shape batch_size x L
            strs: list of strings
            tokens: LongTensor of shape batch_size x L
            padding_mask: ByteTensor of shape batch_size x L
        """
        self.alphabet.cls_idx = self.alphabet.get_idx("<cath>")
        batch = []
        for coords, confidence, seq in raw_batch:
            if confidence is None:
                confidence = 1.0
            if isinstance(confidence, float) or isinstance(confidence, int):
                confidence = [float(confidence)] * len(coords)
            if seq is None:
                seq = "X" * len(coords)
            batch.append(((coords, confidence), seq))

        coords_and_confidence, strs, tokens = super().__call__(batch)

        # pad beginning and end of each protein due to legacy reasons
        coords = [
            F.pad(torch.tensor(cd), (0, 0, 0, 0, 1, 1), value=np.inf)
            for cd, _ in coords_and_confidence
        ]
        confidence = [
            F.pad(torch.tensor(cf), (1, 1), value=-1.0)
            for _, cf in coords_and_confidence
        ]
        coords = self.collate_dense_tensors(coords, pad_v=np.nan)
        confidence = self.collate_dense_tensors(confidence, pad_v=-1.0)
        if device is not None:
            coords = coords.to(device)
            confidence = confidence.to(device)
            tokens = tokens.to(device)
        padding_mask = torch.isnan(coords[:, :, 0, 0])
        coord_mask = torch.isfinite(coords.sum(-2).sum(-1))
        confidence = confidence * coord_mask + (-1.0) * padding_mask
        return coords, confidence, strs, tokens, padding_mask

    def from_lists(self, coords_list, confidence_list=None, seq_list=None, device=None):
        """
        Args:
            coords_list: list of length batch_size, each item is a list of
            floats in shape L x 3 x 3 to describe a backbone
            confidence_list: one of
                - None, default to highest confidence
                - list of length batch_size, each item is a scalar
                - list of length batch_size, each item is a list of floats of
                    length L to describe the confidence scores for the backbone
                    with values between 0. and 1.
            seq_list: either None or a list of strings
        Returns:
            coords: Tensor of shape batch_size x L x 3 x 3
            confidence: Tensor of shape batch_size x L
            strs: list of strings
            tokens: LongTensor of shape batch_size x L
            padding_mask: ByteTensor of shape batch_size x L
        """
        batch_size = len(coords_list)
        if confidence_list is None:
            confidence_list = [None] * batch_size
        if seq_list is None:
            seq_list = [None] * batch_size
        raw_batch = zip(coords_list, confidence_list, seq_list)
        return self.__call__(raw_batch, device)

    @staticmethod
    def collate_dense_tensors(samples, pad_v):
        """
        Takes a list of tensors with the following dimensions:
            [(d_11,       ...,           d_1K),
             (d_21,       ...,           d_2K),
             ...,
             (d_N1,       ...,           d_NK)]
        and stack + pads them into a single tensor of:
        (N, max_i=1,N { d_i1 }, ..., max_i=1,N {diK})
        """
        if len(samples) == 0:
            return torch.Tensor()
        if len(set(x.dim() for x in samples)) != 1:
            raise RuntimeError(
                f"Samples has varying dimensions: {[x.dim() for x in samples]}"
            )
        (device,) = tuple(set(x.device for x in samples))  # assumes all on same device
        max_shape = [max(lst) for lst in zip(*[x.shape for x in samples])]
        result = torch.empty(
            len(samples), *max_shape, dtype=samples[0].dtype, device=device
        )
        result.fill_(pad_v)
        for i in range(len(samples)):
            result_i = result[i]
            t = samples[i]
            result_i[tuple(slice(0, k) for k in t.shape)] = t
        return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from argparse import Namespace

import torch
import torch.nn as nn
import torch.nn.functional as F

from .features import GVPGraphEmbedding
from .gvp_modules import GVPConvLayer, LayerNorm
from .gvp_utils import unflatten_graph


class GVPEncoder(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.embed_graph = GVPGraphEmbedding(args)

        node_hidden_dim = (args.node_hidden_dim_scalar, args.node_hidden_dim_vector)
        edge_hidden_dim = (args.edge_hidden_dim_scalar, args.edge_hidden_dim_vector)

        conv_activations = (F.relu, torch.sigmoid)
        self.encoder_layers = nn.ModuleList(
            GVPConvLayer(
                node_hidden_dim,
                edge_hidden_dim,
                drop_rate=args.dropout,
                vector_gate=True,
                attention_heads=0,
                n_message=3,
                conv_activations=conv_activations,
                n_edge_gvps=0,
                eps=1e-4,
                layernorm=True,
            )
            for i in range(args.num_encoder_layers)
        )

    def forward(self, coords, coord_mask, padding_mask, confidence):
        node_embeddings, edge_embeddings, edge_index = self.embed_graph(
            coords, coord_mask, padding_mask, confidence
        )

        for i, layer in enumerate(self.encoder_layers):
            node_embeddings, edge_embeddings = layer(
                node_embeddings, edge_index, edge_embeddings
            )

        node_embeddings = unflatten_graph(node_embeddings, coords.shape[0])
        return node_embeddings



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

import torch
import torch.nn.functional as F
from scipy.spatial import transform
from torch import Tensor, nn

from antifold.esm.data import Alphabet

from .features import DihedralFeatures
from .gvp_encoder import GVPEncoder
from .gvp_transformer_encoder import GVPTransformerEncoder
from .gvp_utils import unflatten_graph
from .transformer_decoder import TransformerDecoder
from .util import CoordBatchConverter, rotate


class GVPTransformerModel(nn.Module):
    """
    GVP-Transformer inverse folding model.

    Architecture: Geometric GVP-GNN as initial layers, followed by
    sequence-to-sequence Transformer encoder and decoder.
    """

    def __init__(self, args, alphabet):
        super().__init__()
        encoder_embed_tokens = self.build_embedding(
            args,
            alphabet,
            args.encoder_embed_dim,
        )
        decoder_embed_tokens = self.build_embedding(
            args,
            alphabet,
            args.decoder_embed_dim,
        )
        encoder = self.build_encoder(args, alphabet, encoder_embed_tokens)
        decoder = self.build_decoder(args, alphabet, decoder_embed_tokens)
        self.args = args
        self.encoder = encoder
        self.decoder = decoder

    @classmethod
    def build_encoder(cls, args, src_dict, embed_tokens):
        encoder = GVPTransformerEncoder(args, src_dict, embed_tokens)
        return encoder

    @classmethod
    def build_decoder(cls, args, tgt_dict, embed_tokens):
        decoder = TransformerDecoder(
            args,
            tgt_dict,
            embed_tokens,
        )
        return decoder

    @classmethod
    def build_embedding(cls, args, dictionary, embed_dim):
        num_embeddings = len(dictionary)
        padding_idx = dictionary.padding_idx
        emb = nn.Embedding(num_embeddings, embed_dim, padding_idx)
        nn.init.normal_(emb.weight, mean=0, std=embed_dim**-0.5)
        nn.init.constant_(emb.weight[padding_idx], 0)
        return emb

    def forward(
        self,
        coords,
        padding_mask,
        confidence,
        prev_output_tokens,
        return_all_hiddens: bool = False,
        features_only: bool = False,
    ):
        encoder_out = self.encoder(
            coords, padding_mask, confidence, return_all_hiddens=return_all_hiddens
        )
        logits, extra = self.decoder(
            prev_output_tokens,
            encoder_out=encoder_out,
            features_only=features_only,
            return_all_hiddens=return_all_hiddens,
        )
        return logits, extra

    def sample(
        self, coords, partial_seq=None, temperature=1.0, confidence=None, device=None
    ):
        """
        Samples sequences based on multinomial sampling (no beam search).

        Args:
            coords: L x 3 x 3 list representing one backbone
            partial_seq: Optional, partial sequence with mask tokens if part of
                the sequence is known
            temperature: sampling temperature, use low temperature for higher
                sequence recovery and high temperature for higher diversity
            confidence: optional length L list of confidence scores for coordinates
        """
        L = len(coords)
        # Convert to batch format
        batch_converter = CoordBatchConverter(self.decoder.dictionary)
        batch_coords, confidence, _, _, padding_mask = batch_converter(
            [(coords, confidence, None)], device=device
        )

        # Start with prepend token
        mask_idx = self.decoder.dictionary.get_idx("<mask>")
        sampled_tokens = torch.full((1, 1 + L), mask_idx, dtype=int)
        sampled_tokens[0, 0] = self.decoder.dictionary.get_idx("<cath>")
        if partial_seq is not None:
            for i, c in enumerate(partial_seq):
                sampled_tokens[0, i + 1] = self.decoder.dictionary.get_idx(c)

        # Save incremental states for faster sampling
        incremental_state = dict()

        # Run encoder only once
        encoder_out = self.encoder(batch_coords, padding_mask, confidence)

        # Make sure all tensors are on the same device if a GPU is present
        if device:
            sampled_tokens = sampled_tokens.to(device)

        # Decode one token at a time
        for i in range(1, L + 1):
            logits, _ = self.decoder(
                sampled_tokens[:, :i],
                encoder_out,
                incremental_state=incremental_state,
            )
            logits = logits[0].transpose(0, 1)
            logits /= temperature
            probs = F.softmax(logits, dim=-1)
            if sampled_tokens[0, i] == mask_idx:
                sampled_tokens[:, i] = torch.multinomial(probs, 1).squeeze(-1)
        sampled_seq = sampled_tokens[0, 1:]

        # Convert back to string via lookup
        return "".join([self.decoder.dictionary.get_tok(a) for a in sampled_seq])



import logging

log = logging.getLogger(__name__)

import sys

sys.path.insert(0, ".")

import os
from typing import Tuple

import numpy as np
import pandas as pd
import torch

from antifold.esm_multichain_util_custom import (concatenate_coords_any,
                                                 concatenate_coords_HL,
                                                 load_complex_coords)


class InverseData(torch.utils.data.Dataset):
    """
    Prepare dataset for ESM-IF1, including span masking and adding gaussian noise, returning
        batches with coords, confidence, strs, tokens, padding_mask
    Returns: coords, confidence, seq, res_pos, loss_mask, targets
    """

    def __init__(
        self,
        verbose: int = 0,
        loss_mask_flag: bool = False,
        gaussian_noise_flag: bool = True,
        gaussian_scale_A: float = 0.1,
        custom_chain_mode: bool = False,
    ):
        # Variables
        self.loss_mask_flag = loss_mask_flag
        self.gaussian_noise_flag = gaussian_noise_flag
        self.gaussian_scale_A = gaussian_scale_A
        self.seq_pdb_fasta_mismatches = 0
        self.custom_chain_mode = custom_chain_mode
        self.verbose = verbose

        if self.custom_chain_mode:
            log.info(
                "NOTE: Custom chain mode enabled. Will run specified chain(s)."
            )

    def load_coords_HL(
        self, pdb_path: str, Hchain: str, Lchain: str
    ) -> Tuple[np.array, str, np.array]:
        """Read pdb file and extract coordinates of backbone (N, CA, C) atoms of given chains
        Args:
            file path: path to pdb file
            Hchain: heavy chain
            Lchain: light chain
        Returns:
            coords_concatenated: 3d backbone coordinates of extracted structure w/ padding - padded coords set to inf
            seq_concatenated: AA sequence w/ padding
            pos_concatenated: residue positions w/ padding
        """

        # coords, seq = esm.inverse_folding.util.load_coords(fpath=pdb_path, chain=chain)
        coords_dict, seq_dict, pos_dict, posinschain_dict = load_complex_coords(
            pdb_path, [Hchain, Lchain]
        )
        (
            coords_concatenated,
            seq_concatenated,
            pos_concatenated,
            posinschain_concatenated,
        ) = concatenate_coords_HL(
            coords_dict, seq_dict, pos_dict, posinschain_dict, heavy_chain_id=Hchain
        )

        # Limit to IMGT VH/VL regions (pos 1-128)

        return (
            coords_concatenated,
            seq_concatenated,
            pos_concatenated,
            posinschain_concatenated,
        )

    def load_coords_any(
        self, pdb_path: str, chains: list,
    ) -> Tuple[np.array, str, np.array]:
        """Read pdb file and extract coordinates of backbone (N, CA, C) atoms of given chains
        Args:
            file path: path to pdb file
            chains: List of chains ids to extract
        Returns:
            coords_concatenated: 3d backbone coordinates of extracted structure w/ padding - padded coords set to inf
            seq_concatenated: AA sequence w/ padding
            pos_concatenated: residue positions w/ padding
        """

        # Get all chains
        coords_dict, seq_dict, pos_dict, posinschain_dict = load_complex_coords(
            pdb_path, chains,
        )

        # Concatenate
        (
            coords_concatenated,
            seq_concatenated,
            pos_concatenated,
            posinschain_concatenated,
        ) = concatenate_coords_any(
            coords_dict, seq_dict, pos_dict, posinschain_dict, chains,
        )

        # Limit to IMGT VH/VL regions (pos 1-128)

        return (
            coords_concatenated,
            seq_concatenated,
            pos_concatenated,
            posinschain_concatenated,
        )

    def add_gaussian_noise(self, coords: np.array, scale=0.1):
        """Add Gaussian noise at scale 0.1A to each coordinate
        Args:
            coords: 3d backbone coordinates of full_structure (as formatted by load_coords, can be span-masked)
        Returns:
            coords with Gaussian noise added
        """
        return coords + np.random.normal(scale=scale, size=coords.shape)

    def populate(
        self,
        pdbs_csv_or_dataframe: 'str, "path or pd.DataFrame"',
        pdb_dir: str,
        verbose: int = 1,
    ):
        """
        Gets the actual PDB paths to be used for training and testing,
        will filter on the PDBs present in the paragraph CSV dict if set.

        Args:
            pdbs_csv_or_dataframe: path to csv file containing pdb, Hchain and Lchain
        """

        # Accept DataFrame or CSV path
        if isinstance(pdbs_csv_or_dataframe, pd.DataFrame):
            df = pdbs_csv_or_dataframe
            log.info(f"Reading in ({len(df)}) PDBs from DataFrame")
        else:
            if not os.path.exists(pdbs_csv_or_dataframe):
                log.error(
                    f"Unable to find pdbs_csv_or_dataframe {pdbs_csv_or_dataframe}"
                )
                sys.exit(1)

            df = pd.read_csv(pdbs_csv_or_dataframe)
            log.info(f"Populating {len(df)} PDBs from {pdbs_csv_or_dataframe}")

        if not len(df) >= 1:
            log.error(f"CSV file {pdbs_csv_or_dataframe} must contain at least 1 PDB")
            sys.exit(1)

        if (
            not df.columns.isin(["pdb", "Hchain", "Lchain"]).sum() >= 3
            and not self.custom_chain_mode
        ):
            log.error(
                f"CSV file requires columns 'pdb, Hchain, Lchain': found {df.columns}"
            )
            sys.exit(1)

        # Create list of PDB paths and check that they exist
        pdb_path_list = []
        for _pdb in df["pdb"]:
            pdb_path = f"{pdb_dir}/{_pdb}.pdb"

            # Check for PDB/CIF
            pdb_path = pdb_path if os.path.exists(pdb_path) else f"{pdb_dir}/{_pdb}.cif"
            pdb_path_list.append(pdb_path)

            if not os.path.exists(pdb_path):
                raise Exception(f"Unable to find PDB/CIF file: {pdb_path}")

        # Infer order of chain from CSV columns (first item pdb, then chains)
        # Should be Hchain, Lchain, then any order
        chain_order = [c for c in df.columns[1:] if "chain" in c]

        # Construct info dict
        self.pdb_info_dict = {}

        for i in range(len(df)):
            _pdb = df.loc[i, "pdb"]
            _pdb_path = pdb_path_list[i]

            # Extract chains present in CSV
            chains = df.loc[i, chain_order]
            chains = [c for c in chains if not pd.isna(c)]
            if len(chains) == 0:
                raise Exception(f"No chains found for PDB {_pdb}")

            # Infer heavy and light chain
            if len(chains) >= 2:
                Hchain, Lchain = chains[0], chains[1]
            # Infer nanobody
            else:
                Hchain = chains[0]
                Lchain = None

            # PDB name with chains
            _pdb_chainsname = _pdb + "_" + "".join(chains)

            self.pdb_info_dict[i] = {
                "pdb": _pdb,
                "pdb_path": _pdb_path,
                "chain_order": chains,
                "pdb_chainsname": _pdb_chainsname,
                "Hchain": Hchain,
                "Lchain": Lchain,
            }

    def __getitem__(self, idx: int):
        """
        # obtain pdb_info for entry with index idx - pdb_info contains the pdb_path (generated in populate)
        Format data to pass to PyTorch DataLoader (with collate_fn = util.CoordBatchConverter)
        """

        # Experimental - use any chains (e.g. with antigen or for ESM-IF1)
        if self.custom_chain_mode:
            pdb_path = self.pdb_info_dict[idx]["pdb_path"]
            chains = self.pdb_info_dict[idx]["chain_order"]

            coords, seq_pdb, pos_pdb, pos_pdb_arr_str = self.load_coords_any(
                pdb_path, chains
            )

        # Regular mode
        else:
            pdb_path = self.pdb_info_dict[idx]["pdb_path"]
            Hchain = self.pdb_info_dict[idx]["Hchain"]
            Lchain = self.pdb_info_dict[idx]["Lchain"]

            coords, seq_pdb, pos_pdb, pos_pdb_arr_str = self.load_coords_HL(
                pdb_path, Hchain, Lchain,
            )

        # Not used, included for legacy reasons
        targets = np.full(len(pos_pdb_arr_str), np.nan)

        # Add (0.1 Å) gaussian noise to Ca, C, N co-ordinates
        if self.gaussian_noise_flag:
            coords = self.add_gaussian_noise(coords=coords, scale=self.gaussian_scale_A)

        # Initialize empty loss_mask
        loss_mask = np.full(len(coords), fill_value=False)

        # reset padding coords between H and L to np.nan (instead of np.inf, which was set so as not to interfere with masking)
        coords[pos_pdb_arr_str == "nan"] = np.nan

        # confidence currently not taken - update
        confidence = None

        # check masking done correctly - no masking in linker region (between heavy and light chains)
        assert loss_mask[np.where(pos_pdb_arr_str == "nan")].sum() == 0

        return coords, confidence, seq_pdb, pos_pdb, pos_pdb_arr_str, loss_mask, targets

    def __len__(self):
        """Get number of entries in dataset"""
        return len(self.pdb_info_dict)



import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


class SinusoidalPositionalEmbedding(nn.Module):
    def __init__(self, embed_dim, padding_idx, learned=False):
        super().__init__()
        self.embed_dim = embed_dim
        self.padding_idx = padding_idx
        self.register_buffer("_float_tensor", torch.FloatTensor(1))
        self.weights = None

    def forward(self, x):
        bsz, seq_len = x.shape
        max_pos = self.padding_idx + 1 + seq_len
        if self.weights is None or max_pos > self.weights.size(0):
            self.weights = self.get_embedding(max_pos)
        self.weights = self.weights.type_as(self._float_tensor)

        positions = self.make_positions(x)
        return (
            self.weights.index_select(0, positions.view(-1))
            .view(bsz, seq_len, -1)
            .detach()
        )

    def make_positions(self, x):
        mask = x.ne(self.padding_idx)
        range_buf = (
            torch.arange(x.size(1), device=x.device).expand_as(x) + self.padding_idx + 1
        )
        positions = range_buf.expand_as(x)
        return positions * mask.long() + self.padding_idx * (1 - mask.long())

    def get_embedding(self, num_embeddings):
        half_dim = self.embed_dim // 2
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, dtype=torch.float) * -emb)
        emb = torch.arange(num_embeddings, dtype=torch.float).unsqueeze(
            1
        ) * emb.unsqueeze(0)
        emb = torch.cat([torch.sin(emb), torch.cos(emb)], dim=1).view(
            num_embeddings, -1
        )
        if self.embed_dim % 2 == 1:
            # zero pad
            emb = torch.cat([emb, torch.zeros(num_embeddings, 1)], dim=1)
        if self.padding_idx is not None:
            emb[self.padding_idx, :] = 0
        return emb



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup
from pathlib import Path

# Read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding='utf-8')

REQUIREMENTS = [i.strip() for i in open("requirements.txt").readlines()]

setup(
    name="antifold",
    version="0.3.1",
    packages=find_packages(),
    description="Inverse folding of antibodies",
    url="https://github.com/oxpig/AntiFold/",
    author="Magnus Haraldson Høie & Alissa Hummer",
    author_email="maghoi@dtu.dk",
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=REQUIREMENTS,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.10",
    ],
)



import glob
import logging
import os
import sys
import warnings
import urllib.request
from pathlib import Path

ROOT_PATH = Path(os.path.dirname(__file__)).parent
sys.path.insert(0, ROOT_PATH)

import re
from collections import OrderedDict

import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F
from Bio import SeqIO
from Bio.Seq import Seq
from biotite.structure.io import pdb

import antifold.esm
from antifold.esm_util_custom import CoordBatchConverter_mask_gpu
from antifold.if1_dataset import InverseData

log = logging.getLogger(__name__)

amino_list = list("ACDEFGHIKLMNPQRSTVWY")

IMGT_dict = {
    "all": range(1, 128 + 1),
    "allH": range(1, 128 + 1),
    "allL": range(1, 128 + 1),
    "FWH": list(range(1, 26 + 1)) + list(range(40, 55 + 1)) + list(range(66, 104 + 1)),
    "FWL": list(range(1, 26 + 1)) + list(range(40, 55 + 1)) + list(range(66, 104 + 1)),
    "CDRH": list(range(27, 39)) + list(range(56, 65 + 1)) + list(range(105, 117 + 1)),
    "CDRL": list(range(27, 39)) + list(range(56, 65 + 1)) + list(range(105, 117 + 1)),
    "FW1": range(1, 26 + 1),
    "FWH1": range(1, 26 + 1),
    "FWL1": range(1, 26 + 1),
    "CDR1": range(27, 39),
    "CDRH1": range(27, 39),
    "CDRL1": range(27, 39),
    "FW2": range(40, 55 + 1),
    "FWH2": range(40, 55 + 1),
    "FWL2": range(40, 55 + 1),
    "CDR2": range(56, 65 + 1),
    "CDRH2": range(56, 65 + 1),
    "CDRL2": range(56, 65 + 1),
    "FW3": range(66, 104 + 1),
    "FWH3": range(66, 104 + 1),
    "FWL3": range(66, 104 + 1),
    "CDR3": range(105, 117 + 1),
    "CDRH3": range(105, 117 + 1),
    "CDRL3": range(105, 117 + 1),
    "FW4": range(118, 128 + 1),
    "FWH4": range(118, 128 + 1),
    "FWL4": range(118, 128 + 1),
}


def extract_chains_biotite(pdb_file):
    """Extract chains in order"""
    pdbf = pdb.PDBFile.read(pdb_file)
    structure = pdb.get_structure(pdbf, model=1)
    return pd.unique(structure.chain_id)


def generate_pdbs_csv(pdbs_dir, max_chains=10):
    """Generates AntiFOld CSV with PDB/CIF basenames + chains (up to 10)"""

    columns = [
        "pdb",
        "Hchain",
        "Lchain",
        "chain3",
        "chain4",
        "chain5",
        "chain6",
        "chain7",
        "chain8",
        "chain9",
        "chain10",
    ]
    pdbs_csv = pd.DataFrame(columns=columns)
    pdbs = glob.glob(f"{pdbs_dir}/*.[pdb cif]*")

    for i, pdb_file in enumerate(pdbs):
        _pdb = os.path.splitext(os.path.basename(pdb_file))[0]

        # Extract first 10 chains
        chains = extract_chains_biotite(pdb_file)[:max_chains]

        # Put pdb and chains in dataframe
        row = [_pdb] + list(chains)
        _cols = columns[: len(row)]
        pdbs_csv.loc[i, _cols] = row

    return pdbs_csv


def load_IF1_checkpoint(model, checkpoint_path: str = ""):
    # Load
    log.debug(f"Loading AntiFold model {checkpoint_path} ...")

    # Check for CPU/GPU load
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    checkpoint_dict = torch.load(checkpoint_path, map_location=torch.device(device))

    # PYL checkpoint ?
    if "model_state_dict" in checkpoint_dict.keys():
        pretrained_dict = {
            re.sub("model.", "", k): v
            for k, v in checkpoint_dict["model_state_dict"].items()
        }

    # PYL checkpoint ?
    elif "state_dict" in checkpoint_dict.keys():
        # Remove "model." from keys
        pretrained_dict = {
            re.sub("model.", "", k): v for k, v in checkpoint_dict["state_dict"].items()
        }

    # IF1-raw?
    else:
        pretrained_dict = checkpoint_dict

    # Load pretrained weights
    model.load_state_dict(pretrained_dict)

    return model


def load_model(checkpoint_path: str = ""):
    """Load raw/FT IF1 model"""

    # Check that AntiFold weights are downloaded
    root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    model_path = f"{root_dir}/models/model.pt"

    if not os.path.exists(model_path):
        log.warning(
            f"Downloading AntiFold model weights from https://opig.stats.ox.ac.uk/data/downloads/AntiFold/models/model.pt to {model_path}"
        )
        url = "https://opig.stats.ox.ac.uk/data/downloads/AntiFold/models/model.pt"
        filename = model_path

        os.makedirs(f"{root_dir}/models", exist_ok=True)
        urllib.request.urlretrieve(url, filename)

    if not os.path.exists(model_path) and not checkpoint_path == "ESM-IF1":
        raise Exception(
            f"Unable to find model weights. File does not exist: {checkpoint_path}"
        )

    # Download IF1 weights
    if checkpoint_path == "ESM-IF1":
        log.info(
            f"NOTE: Loading ESM-IF1 weights instead of fine-tuned AntiFold weights"
        )
        # Suppress regression weights warning - not needed
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model, _ = antifold.esm.pretrained.esm_if1_gvp4_t16_142M_UR50()

    # Load AntiFold weights locally
    else:
        model, _ = antifold.esm.pretrained._load_IF1_local()
        model = load_IF1_checkpoint(model, model_path)

    # Evaluation mode when predicting
    model = model.eval()

    # Send to CPU/GPU
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    _ = model.to(device)
    log.info(f"Loaded model to {device}.")

    return model


def get_dataset_pdb_name_chainsname_res_posins_chains(dataset, idx):
    """Gets PDB sequence, position+insertion codes and chains from dataset"""

    # PDB path
    pdb_path = dataset.pdb_info_dict[idx]["pdb_path"]

    # PDB names
    pdb_name = dataset.pdb_info_dict[idx]["pdb"]
    pdb_chainsname = dataset.pdb_info_dict[idx]["pdb_chainsname"]

    # Sequence - gaps
    seq = dataset[idx][2]
    pdb_res = [aa for aa in seq if aa != "-"]

    # Position + insertion codes - gaps
    posinschain = dataset[idx][4]
    posinschain = [p for p in posinschain if p != "nan"]

    # Extract position + insertion code + chain (1-letter)
    pdb_posins = [p[:-1] for p in posinschain]
    pdb_chains = [p[-1] for p in posinschain]

    return pdb_name, pdb_chainsname, pdb_res, pdb_posins, pdb_chains


def logits_to_seqprobs_list(logits, tokens):
    """Convert logits (bs x 35 x L) ot list of L x 20 seqprobs"""

    alphabet = antifold.esm.data.Alphabet.from_architecture("invariant_gvp")

    mask_gap = tokens[:, 1:] != 30  # 30 is gap
    mask_pad = tokens[:, 1:] != alphabet.padding_idx  # 1 is padding
    mask_combined = mask_gap & mask_pad

    # Check that only 10x gap ("-") per sequence!
    # batch_size = logits.shape[0]
    # assert (mask_gap == False).sum() == batch_size * 10

    # Filter out gap (-) and padding (<pad>) positions, only keep 21x amino-acid probs (4:24) + "X"
    seqprobs_list = [logits[i, 4:25, mask_combined[i]] for i in range(len(logits))]

    return seqprobs_list


def get_dataset_dataloader(
    pdbs_csv_or_dataframe, pdb_dir, batch_size, custom_chain_mode=False, num_threads=0
):
    """Prepares dataset/dataoader from CSV file containing PDB paths and H/L chains"""

    # Set number of threads & workers
    if num_threads >= 1:
        torch.set_num_threads(num_threads)
        num_threads = min(num_threads, 4)

    # Load PDB coordinates
    dataset = InverseData(
        gaussian_noise_flag=False, custom_chain_mode=custom_chain_mode,
    )
    dataset.populate(pdbs_csv_or_dataframe, pdb_dir)

    # Prepare torch dataloader at specified batch size
    alphabet = antifold.esm.data.Alphabet.from_architecture("invariant_gvp")
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        collate_fn=CoordBatchConverter_mask_gpu(alphabet),
        num_workers=num_threads,
    )

    return dataset, dataloader


def dataset_dataloader_to_predictions_list(
    model, dataset, dataloader, batch_size=1, extract_embeddings=False
):
    """Get PDB predictions from a dataloader"""

    # Check that dataloader and dataset match, and no random shuffling
    if "random" in str(dataloader.sampler).lower():
        raise ValueError(
            "Torch DataLoader sampler must not be random. Did you forget to set torch.utils.data.DataLoader ... shuffle=False?"
        )
    if dataloader.dataset is not dataset:
        raise ValueError("Dataloader and dataset must match to align samples!")

    # Get all batch predictions
    all_seqprobs_list = []
    all_embeddings_list = []
    for bi, batch in enumerate(dataloader):
        start_index = bi * batch_size
        end_index = min(
            start_index + batch_size, len(dataset)
        )  # Adjust for the last batch
        log.info(
            f"Predicting batch {bi+1}/{len(dataloader)}: PDBs {start_index+1}-{end_index} out of {len(dataset)} total"
        )  # -1 because the end_index is exclusive

        # Test dataloader
        (
            coords,
            confidence,
            strs,
            tokens,
            padding_mask,
            loss_masks,
            res_pos,
            posinschain_list,
            targets,
        ) = batch

        # Test forward
        with torch.no_grad():
            prev_output_tokens = tokens[:, :-1]
            logits, extra = model.forward(  # bs x 35 x seq_len, exlude bos, eos
                coords,
                padding_mask,  # Includes masked positions
                confidence,
                prev_output_tokens,
                features_only=False,
            )

            logits = logits.detach().cpu().numpy()
            tokens = tokens.detach().cpu().numpy()

            # List of L x 21 seqprobs (20x AA, 21st == "X")
            L = logits_to_seqprobs_list(logits, tokens)
            all_seqprobs_list.extend(L)

            if extract_embeddings:
                # Extract embeddings from batch
                padding_mask = padding_mask.detach().cpu().numpy()  # bs x len
                Es = extra["inner_states"][0].detach().cpu().numpy()  # len x bs x 512
                # L[0] = encoder output
                # L[1] = decoder input
                # L[2] = decoder output layer 1
                # L[3] = decoder output layer 2 ...
                Es = np.transpose(Es, (1, 0, 2))  # bs x len x 512

                # Embeddings without padding and bos/eos
                for e, pad in zip(Es, padding_mask):
                    # e: len x 512
                    e = e[~pad]  # (len - pad) x 512
                    e = e[1:-1]  # seq_len x 512
                    all_embeddings_list.append(e)  # seq_len x 512

    return all_seqprobs_list, all_embeddings_list


def predictions_list_to_df_logits_list(all_seqprobs_list, dataset, dataloader):
    """PDB preds list to PDB DataFrames"""

    # Check that dataloader and dataset match, and no random shuffling
    if "random" in str(dataloader.sampler).lower():
        raise ValueError(
            "Torch DataLoader sampler must not be random. Did you forget to set torch.utils.data.DataLoader ... shuffle=False?"
        )
    if dataloader.dataset is not dataset:
        raise ValueError("Dataloader and dataset must match to align samples!")

    # Create DataFrame for each PDB
    all_df_logits_list = []
    for idx, seq_probs in enumerate(all_seqprobs_list):
        # Get PDB sequence, position+insertion code and H+L chain idxs
        (
            pdb_name,
            pdb_chainsname,
            pdb_res,
            pdb_posins,
            pdb_chains,
        ) = get_dataset_pdb_name_chainsname_res_posins_chains(dataset, idx)

        # Check matches w/ residue probs
        assert len(seq_probs) == len(pdb_posins)

        # Logits to DataFrame
        alphabet = antifold.esm.data.Alphabet.from_architecture("invariant_gvp")
        df_logits = pd.DataFrame(data=seq_probs, columns=alphabet.all_toks[4:25],)

        # Limit to 20x amino-acids probs
        _alphabet = list("ACDEFGHIKLMNPQRSTVWY")
        df_logits = df_logits[_alphabet]

        # Add PDB info
        positions = pdb_posins_to_pos(pd.Series(pdb_posins))
        top_res = np.array((_alphabet))[df_logits[_alphabet].values.argmax(axis=1)]
        perplexity = calc_pos_perplexity(df_logits)

        # Add to DataFrame
        df_logits.name = pdb_chainsname
        df_logits.insert(0, "pdb_posins", pdb_posins)
        df_logits.insert(1, "pdb_chain", pdb_chains)
        df_logits.insert(2, "pdb_res", pdb_res)
        df_logits.insert(3, "top_res", top_res)
        df_logits.insert(4, "pdb_pos", positions)
        df_logits.insert(5, "perplexity", perplexity)

        # Skip if not IMGT numbered - 10 never found in H-chain IMGT numbered PDBs
        Hchain = pdb_chains[0]
        Hpos = positions[pdb_chains == Hchain]
        if 10 in Hpos and not dataset.custom_chain_mode:
            log.error(
                f"WARNING: PDB {pdb_name} seems to not be IMGT numbered! Output probabilities may be affected. See https://opig.stats.ox.ac.uk/webapps/sabdab-sabpred/sabpred/anarci/"
            )
        # Limit to IMGT positions only (only ones trained on)
        # imgt_mask = get_imgt_mask(df_logits, imgt_regions=["all"])
        # df_logits = df_logits[imgt_mask]

        all_df_logits_list.append(df_logits)

    return all_df_logits_list


def df_logits_list_to_logprob_csvs(
    df_logits_list, out_dir, embeddings_list=False, float_format="%.4f"
):
    """Save df_logits_list to CSVs"""
    os.makedirs(out_dir, exist_ok=True)
    log.info(f"Saving {len(df_logits_list)} CSVs to {out_dir}")

    for i, df in enumerate(df_logits_list):
        # Convert to log-probs
        df_out = df_logits_to_logprobs(df)
        # Save
        outpath = f"{out_dir}/{df.name}.csv"
        log.info(f"Writing {df.name} per-residue log probs CSV to {outpath}")
        df_out.to_csv(outpath, float_format=float_format, index=False)

        if embeddings_list:
            # Save embeddingsl
            outpath = f"{out_dir}/{df.name}.npy"
            log.info(f"Writing {df.name} per-residue embeddings to {outpath}")
            np.save(outpath, embeddings_list[i])


def seed_everything(seed: int):
    # https://gist.github.com/ihoromi4/b681a9088f348942b01711f251e5f964
    import os
    import random

    import numpy as np
    import torch

    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True


def get_pdbs_logits(
    model,
    pdbs_csv_or_dataframe,
    pdb_dir,
    out_dir=False,
    batch_size=1,
    extract_embeddings=False,
    custom_chain_mode=False,
    num_threads=0,
    save_flag=False,
    float_format="%.4f",
    seed=42,
):
    """Predict PDBs from a CSV file"""

    # if save_flag:
    #    log.info(f"Saving prediction CSVs to {out_dir}")

    seed_everything(seed)

    # Load PDBs
    dataset, dataloader = get_dataset_dataloader(
        pdbs_csv_or_dataframe,
        pdb_dir,
        batch_size=batch_size,
        custom_chain_mode=custom_chain_mode,
        num_threads=num_threads,
    )

    # Predict PDBs -> df_logits
    predictions_list, embeddings_list = dataset_dataloader_to_predictions_list(
        model,
        dataset,
        dataloader,
        batch_size=batch_size,
        extract_embeddings=extract_embeddings,
    )
    df_logits_list = predictions_list_to_df_logits_list(
        predictions_list, dataset, dataloader
    )

    # Save df_logits to CSVs
    if save_flag:
        df_logits_list_to_logprob_csvs(
            df_logits_list,
            out_dir,
            embeddings_list=embeddings_list,
            float_format=float_format,
        )

    if extract_embeddings:
        return df_logits_list, embeddings_list
    else:
        return df_logits_list


def calc_pos_perplexity(df):
    cols = list("ACDEFGHIKLMNPQRSTVWY")
    t = torch.tensor(df[cols].values)
    probs = F.softmax(t, dim=1)
    perplexities = torch.pow(2, -(probs * torch.log2(probs)).sum(dim=1))

    return perplexities.numpy()


def sample_new_sequences_CDR_HL(
    df,
    t=0.20,
    imgt_regions=["CDR1", "CDR2", "CDR3"],
    exclude_heavy=False,
    exclude_light=False,
    return_mutation_df=False,
    limit_expected_variation=True,
    verbose=False,
):
    """Samples new sequences only varying at H/L CDRs"""

    def _sample_cdr_seq(df, imgt_regions, t=0.20):
        """DF to sampled seq"""

        # CDR1+2+3 mask
        region_mask = get_imgt_mask(df, imgt_regions)

        # Probabilities after scaling with temp
        probs = get_temp_probs(df, t=t)
        probs_cdr = probs[region_mask]

        # Sampled tokens and sequence
        sampled_tokens = torch.multinomial(probs_cdr, 1).squeeze(-1)
        sampled_seq = np.array([amino_list[i] for i in sampled_tokens])

        return sampled_seq

    # Prepare to sample new H + L sequences
    df_H, df_L = get_dfs_HL(df)

    # Get H, sampling only for (CDR1, 2, 3)
    H_sampled = get_df_seq(df_H)

    regions = [region for region in imgt_regions if "L" not in region]
    if len(regions) > 0 and not exclude_heavy:
        region_mask = get_imgt_mask(df_H, regions)
        H_sampled[region_mask] = _sample_cdr_seq(df_H, regions, t=t)

    # Get L, sampling only for (CDR1, 2, 3)
    L_sampled = get_df_seq(df_L)

    regions = [region for region in imgt_regions if "H" not in region]
    if len(regions) > 0 and not exclude_light:
        region_mask = get_imgt_mask(df_L, regions)
        L_sampled[region_mask] = _sample_cdr_seq(df_L, regions, t=t)

    # Use for later
    sampled_seq = np.concatenate([H_sampled, L_sampled])
    region_mask = get_imgt_mask(df, imgt_regions)

    # Mismatches vs predicted (CDR only)
    pred_seq = get_df_seq_pred(df)
    mismatch_idxs_pred_cdr = np.where(
        (sampled_seq[region_mask] != pred_seq[region_mask])
    )[0]

    # Mismatches vs original (all)
    orig_seq = get_df_seq(df)
    mismatch_idxs_orig = np.where((sampled_seq != orig_seq))[0]

    # Limit mutations (backmutate) to as many expected from temperature sampling
    if limit_expected_variation:
        backmutate = len(mismatch_idxs_orig) - len(mismatch_idxs_pred_cdr)

        if backmutate >= 1:
            backmutate_idxs = np.random.choice(
                mismatch_idxs_orig, size=backmutate, replace=False
            )
            sampled_seq[backmutate_idxs] = orig_seq[backmutate_idxs]
            H_sampled = sampled_seq[: len(df_H)]
            L_sampled = sampled_seq[-len(df_L) :]

    # Variables for calculating mismatches
    sampled_seq = np.concatenate([H_sampled, L_sampled])
    orig_seq = get_df_seq(df)

    # DataFrame with sampled mutations
    if return_mutation_df:
        mut_list = np.where(sampled_seq != orig_seq)[0]
        df_mut = df.loc[
            mut_list, ["pdb_res", "top_res", "pdb_posins", "pdb_chain"]
        ].copy()
        df_mut.insert(1, "aa_sampled", sampled_seq[mut_list])

        return H_sampled, L_sampled, df_mut

    return H_sampled, L_sampled, df


def pdb_posins_to_pos(pdb_posins):
    # Convert pos+insertion code to numerical only
    return pdb_posins.astype(str).str.extract(r"(\d+)")[0].astype(int).values


def get_imgt_mask(df, imgt_regions=["CDR1", "CDR2", "CDR3"]):
    """Returns e.g. CDR1+2+3 mask"""

    positions = pdb_posins_to_pos(df["pdb_posins"])
    region_pos_list = list()

    for region in imgt_regions:
        if str(region) not in IMGT_dict.keys():
            region_pos_list.extend(region)
        else:
            region_pos_list.extend(list(IMGT_dict[region]))

    region_mask = pd.Series(positions).isin(region_pos_list).values

    return region_mask


def get_df_logits(df):
    cols = list("ACDEFGHIKLMNPQRSTVWY")
    logits = torch.tensor(df[cols].values)

    return logits


def get_temp_probs(df, t=0.20):
    """Gets temperature scaled probabilities for sampling"""

    logits = get_df_logits(df)
    temp_logits = logits / np.max([t, 0.001])  # Lower-bound to prevent div by 0
    temp_probs = F.softmax(temp_logits, dim=1)

    return temp_probs


def get_dfs_HL(df):
    """Split df into heavy and light chains"""
    Hchain, Lchain = df["pdb_chain"].unique()[:2] # Assume heavy, light
    return df[df["pdb_chain"] == Hchain], df[df["pdb_chain"] == Lchain]


def get_df_seq(df):
    """Get PDB sequence"""
    return df["pdb_res"].values


def get_df_seq_pred(df):
    """Get PDB sequence"""
    return df["top_res"].values


def get_df_seqs_HL(df):
    """Get heavy and light chain sequences"""
    df_H, df_L = get_dfs_HL(df)
    return get_df_seq(df_H), get_df_seq(df_L)


def sample_from_df_logits(
    df_logits,
    sample_n=1,
    sampling_temp=0.20,
    regions_to_mutate=["CDR1", "CDR2", "CDR3"],
    exclude_heavy=False,
    exclude_light=False,
    limit_expected_variation=False,
    verbose=False,
    seed=42,
):
    # Get original H/L sequence
    H_orig, L_orig = get_df_seqs_HL(df_logits)

    # Only sampling from heavy, light chains
    df_logits_HL = df_logits.iloc[:len(H_orig) + len(L_orig), :]
    df_logits_HL.name = df_logits.name

    # Stats
    seq = "".join(H_orig) + "".join(L_orig)
    _, global_score = get_sequence_sampled_global_score(
        seq, df_logits_HL, regions_to_mutate
    )

    # Save to FASTA dict
    fasta_dict = OrderedDict()
    _id = f"{df_logits_HL.name}"
    desc = f", score={global_score:.4f}, global_score={global_score:.4f}, regions={regions_to_mutate}, model_name=AntiFold, seed={seed}"
    seq = "".join(H_orig) + "/" + "".join(L_orig)
    fasta_dict[_id] = SeqIO.SeqRecord(Seq(seq), id=_id, name="", description=desc)

    if verbose:
        log.info(f"{_id}: {desc}")

    if not isinstance(sampling_temp, list):
        sampling_temp = [sampling_temp]

    for t in sampling_temp:
        # Sample sequences n times
        for n in range(sample_n):
            # Get mutated H/L sequence
            H_mut, L_mut, df_mut = sample_new_sequences_CDR_HL(
                df_logits_HL,  # DataFrame with residue probabilities
                t=t,  # Sampling temperature
                imgt_regions=regions_to_mutate,  # Region to sample
                exclude_heavy=exclude_heavy,  # Allow mutations in heavy chain
                exclude_light=exclude_light,  # Allow mutation in light chain
                limit_expected_variation=limit_expected_variation,  # Only mutate as many positions are expected from temperature
                verbose=verbose,
            )

            # Original sequence
            seq_orig = "".join(H_orig) + "".join(L_orig)

            # Sequence recovery and mismatches
            correct_matches = (H_orig == H_mut).sum() + (L_orig == L_mut).sum()
            seq_recovery = correct_matches / len(seq_orig)
            n_mut = (H_orig != H_mut).sum() + (L_orig != L_mut).sum()

            seq_mut = "".join(H_mut) + "".join(L_mut)
            score_sampled, global_score = get_sequence_sampled_global_score(
                seq_mut, df_logits_HL, regions_to_mutate
            )

            # Save to FASTA dict
            _id = f"{df_logits_HL.name}__{n+1}"
            desc = f"T={t:.2f}, sample={n+1}, score={score_sampled:.4f}, global_score={global_score:.4f}, seq_recovery={seq_recovery:.4f}, mutations={n_mut}"
            seq_mut = "".join(H_mut) + "/" + "".join(L_mut)
            fasta_dict[_id] = SeqIO.SeqRecord(
                Seq(seq_mut), id="", name="", description=desc
            )

            if verbose:
                log.info(f"{_id}: {desc}")

    return fasta_dict


def write_fasta_to_dir(fasta_dict, df_logits, out_dir, verbose=True):
    """Write fasta to output folder"""

    os.makedirs(out_dir, exist_ok=True)
    outfile = f"{out_dir}/{df_logits.name}.fasta"
    if verbose:
        log.info(f"Saving to {outfile}")

    with open(outfile, "w") as out_handle:
        SeqIO.write(fasta_dict.values(), out_handle, "fasta")


def visualize_mutations(orig, mut, chain=""):
    """Visualize mutations between two sequences"""

    # Convert to numpy array
    # (whether string, list, Bio.Seq.Seq or np.array)
    orig = np.array(list(orig))
    mut = np.array(list(mut))
    mismatches = "".join(["X" if match else "_" for match in (orig != mut)])

    # Print
    print(f"Mutations ({(orig != mut).sum()}):\t{mismatches}")
    print(f"Original {chain}:\t\t{''.join(orig)}")
    print(f"Mutated {chain}:\t\t{''.join(mut)}\n")


def df_logits_to_probs(df_logits):
    """Convert logits to probabilities"""

    # Calculate probabilities
    amino_list = list("ACDEFGHIKLMNPQRSTVWY")
    t = torch.tensor(df_logits[amino_list].values)
    probs = F.softmax(t, dim=1)

    # Insert into copied dataframe
    df_probs = df_logits.copy()
    df_probs[amino_list] = probs

    return df_probs


def df_logits_to_logprobs(df_logits):
    """Convert logits to probabilities"""

    # Calculate probabilities
    amino_list = list("ACDEFGHIKLMNPQRSTVWY")
    t = torch.tensor(df_logits[amino_list].values)
    probs = F.log_softmax(t, dim=1)

    # Insert into copied dataframe
    df_probs = df_logits.copy()
    df_probs[amino_list] = probs

    return df_probs


def sequence_to_onehot(sequence):
    amino_list = list("ACDEFGHIKLMNPQRSTVWY")
    one_hot = np.zeros((len(sequence), len(amino_list)), dtype=int)
    for i, res in enumerate(sequence):
        one_hot[i, amino_list.index(res)] = 1
    return one_hot


def get_sequence_sampled_global_score(seq, df_logits, regions_to_mutate=False):
    """
    Get average log probability of sampled / all amino acids
    """

    def _scores(S, log_probs, mask):
        criterion = torch.nn.NLLLoss(reduction="none")
        loss = criterion(
            log_probs.contiguous().view(-1, log_probs.size(-1)), S.contiguous().view(-1)
        ).view(S.size())
        scores = torch.sum(loss * mask, dim=-1) / torch.sum(mask, dim=-1)
        return scores

    # One-hot to indices
    onehot = sequence_to_onehot(seq)
    S = torch.argmax(torch.tensor(onehot, dtype=torch.float), dim=-1)

    # Logits to log probs
    logits = torch.tensor(df_logits[amino_list].values)
    log_probs = F.log_softmax(logits, dim=-1)

    # clip probs
    # log_probs = torch.clamp(log_probs, min=-100, max=100)

    # Calculate log odds scores
    mask = torch.ones_like(S, dtype=torch.bool)
    score_global = _scores(S, log_probs, mask)

    if regions_to_mutate:
        region_mask = get_imgt_mask(df_logits, regions_to_mutate)
        mask = torch.tensor(region_mask)
        score_sampled = _scores(S, log_probs, mask)
    else:
        score_sampled = score_global

    return score_sampled.item(), score_global.item()
